"""
bulk-seq vdj mapping
"""

import subprocess
import csv
import os

from celescope.tools import utils, step
from celescope.vdj import mapping_vdj as super_vdj
from celescope.vdj.__init__ import CHAINS
from celescope.bulk_rna.starsolo import get_barcode_sample, get_well_barcode
import celescope.tools.parse_chemistry as parse_chemistry
from celescope.__init__ import HELP_DICT
from celescope.chemistry_dict import chemistry_dict
import pandas as pd
import numpy as np


def inverse_simpson(counts):
    """
    >>> inverse_simpson([10,20,30])
    2.57
    """
    counts = np.array(counts)
    if np.sum(counts) == 0:
        return 0.0
    proportions = counts / np.sum(counts)
    return round(1.0 / np.sum(proportions**2), 2)


class Mapping_vdj(step.Step):
    """
    ## Features
    - Align R2 reads to IMGT(http://www.imgt.org/) database sequences with blast.
    ## Output
    - `{sample}_airr.tsv` The alignment result of each read.
    A tab-delimited file compliant with the AIRR Rearrangement schema(https://docs.airr-community.org/en/stable/datarep/rearrangements.html)
    - `{sample}_produtive.tsv` Including all productive chains.
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        self.chains = CHAINS[args.type]

        chemistry = self.get_slot_key(
            slot="metrics", step_name="sample", key="Chemistry"
        )

        self.pattern_dict, bc = parse_chemistry.get_pattern_dict_and_bc(
            chemistry, args.pattern, args.whitelist
        )
        self.barcode_sample = get_barcode_sample(bc[0], args.well_sample)
        well_barcode = get_well_barcode(bc[0])
        self.barcode_well = {v: k for k, v in well_barcode.items()}

        # out
        self.annotation_dir = f"{self.outdir}/annotation"
        if not os.path.exists(self.annotation_dir):
            os.makedirs(self.annotation_dir)
        self.clonotypes_dir = f"{self.outdir}/clonotypes"
        if not os.path.exists(self.clonotypes_dir):
            os.makedirs(self.clonotypes_dir)
        self.airr_out = f"{self.out_prefix}_airr.tsv"
        self.annotation_csv = f"{self.out_prefix}_filtered_annotations.csv"
        self.clonotypes_csv = f"{self.out_prefix}_clonotypes.csv"
        self.outs = [
            self.annotation_csv,
            self.clonotypes_csv,
            self.annotation_dir,
            self.clonotypes_dir,
        ]

    @utils.add_log
    def igblast(self):
        if self.args.type == "TCR":
            chain = "TR"
            ig_seqtype = "TCR"
        elif self.args.type == "BCR":
            chain = "IG"
            ig_seqtype = "Ig"
        cmd = (
            f"igblastn -query {self.args.fasta} "
            f"-organism {self.args.species} "
            f"-ig_seqtype {ig_seqtype} "
            f"-auxiliary_data optional_file/{self.args.species}_gl.aux "
            f"-num_threads {self.args.thread} "
            f"-germline_db_V {self.args.ref_path}/{chain}V.fa "
            f"-germline_db_D {self.args.ref_path}/{chain}D.fa "
            f"-germline_db_J {self.args.ref_path}/{chain}J.fa "
            "-domain_system imgt -show_translation -outfmt 19 "  # outfmt19 is an AIRR tab-delimited file, IgBLAST v1.9.0 or higher required.
            f"-out {self.airr_out} "
        )
        self.igblast.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def process_airr(self):
        """
        read airr line by line, collect metrics, write productive to seperate wells
        """

        tsv_handles = {
            barcode: open(f"{self.annotation_dir}/{sample}_annotation.csv", "wt")
            for barcode, sample in self.barcode_sample.items()
        }
        tsv_headers = [
            "barcode",
            "well",
            "sample",
            "umi",
            "chain",
            "v_gene",
            "d_gene",
            "j_gene",
            "c_gene",
            "productive",
            "cdr3",
            "cdr3_nt",
            "umis",
        ]
        for f in tsv_handles.values():
            f.write(",".join(tsv_headers) + "\n")

        consensus_metrics_df = pd.read_csv(
            self.args.consensus_metrics_file, sep="\t", index_col=0
        )
        consensus_metrics_dict = consensus_metrics_df.to_dict(orient="index")
        metrics = utils.nested_defaultdict(dim=2)
        for barcode in consensus_metrics_dict:
            if barcode in self.barcode_sample:
                metrics[barcode]["well"] = self.barcode_well[barcode]
                metrics[barcode]["sample"] = self.barcode_sample[barcode]
            else:
                metrics[barcode]["well"] = ""
                metrics[barcode]["sample"] = ""
            metrics[barcode]["n_clonotypes"] = 0
            metrics[barcode]["diversity"] = 0
            metrics[barcode]["read"] = consensus_metrics_dict[barcode]["n_read"]
            metrics[barcode]["umi"] = 0
            metrics[barcode]["umi_mapped"] = 0
            metrics[barcode]["umi_confident"] = 0
            for chain in self.chains:
                metrics[barcode][f"umi_confident_{chain}"] = 0

        total_umi = 0
        with open(self.airr_out, "rt") as infile:
            reader = csv.DictReader(infile, delimiter="\t")
            for row in reader:
                total_umi += 1
                barcode, umi = row["sequence_id"].split(":")[0:2]
                metrics[barcode]["umi"] += 1
                if row["v_call"] != "" or row["d_call"] != "" or row["j_call"] != "":
                    metrics[barcode]["umi_mapped"] += 1
                else:
                    continue
                if (
                    row["productive"] == "T"
                    and row["junction"] != ""
                    and "N" not in row["junction"]
                ):
                    metrics[barcode]["umi_confident"] += 1
                    if row["locus"] in self.chains:
                        metrics[barcode][f"umi_confident_{row['locus']}"] += 1
                else:
                    continue
                if barcode in self.barcode_sample:
                    for gene in ["v_call", "d_call", "j_call"]:
                        row[gene] = row[gene].split("*")[0]
                    line = [
                        barcode,
                        str(self.barcode_well[barcode]),
                        self.barcode_sample[barcode],
                        umi,
                        row["locus"],
                        row["v_call"],
                        row["d_call"],
                        row["j_call"],
                        "None",
                        "True",
                        row["junction_aa"],
                        row["junction"],
                        "1",  # umis
                    ]
                    tsv_handles[barcode].write(",".join(line) + "\n")

        umi_mapped = sum(v for barcode, k in metrics.items() for v in [k["umi_mapped"]])
        self.add_metric(
            name="UMIs Mapped to Any VDJ Gene",
            value=umi_mapped,
            total=total_umi,
            help_info="UMIs Mapped to any germline VDJ gene segments",
        )

        umi_confident = sum(
            v for barcode, k in metrics.items() for v in [k["umi_confident"]]
        )
        self.add_metric(
            name="UMIs Mapped Confidently to VJ Gene",
            value=umi_confident,
            total=total_umi,
            help_info="UMIs with productive rearrangement mapped to VJ gene pairs and without N in the junction sequence",
        )

        for chain in self.chains:
            umi_confident_chain = sum(
                v
                for barcode, k in metrics.items()
                for v in [k[f"umi_confident_{chain}"]]
            )
            self.add_metric(
                name=f"UMIs Mapped Confidently to {chain}",
                value=umi_confident_chain,
                total=total_umi,
            )

        self.metrics = metrics

    @staticmethod
    def create_well_clonotypes(annotation_file, out_file):
        df_anno = pd.read_csv(annotation_file)
        df_clono = (
            df_anno.groupby(["barcode", "well", "sample", "chain", "cdr3"])
            .size()
            .reset_index(name="umis")
        )
        sample = df_anno["sample"][0]
        df_clono.sort_values("umis", ascending=False, inplace=True)
        df_clono["raw_clonotype_id"] = [
            "{}_{}".format(sample, i + 1) for i in range(len(df_clono))
        ]

        df_anno = df_anno.merge(
            df_clono[
                ["barcode", "well", "sample", "chain", "cdr3", "raw_clonotype_id"]
            ],
            on=["barcode", "well", "sample", "chain", "cdr3"],
            how="left",
        )
        # add umi to treat each umi as a single cell;compatible with immunarch
        df_anno["barcode"] = df_anno["barcode"] + "_" + df_anno["umi"]
        df_anno.to_csv(annotation_file, index=False)

        total_umis = sum(df_clono["umis"])
        df_clono["percent"] = df_clono["umis"].apply(
            lambda x: f"{x / total_umis * 100:.2f}%"
        )

        df_clono.to_csv(out_file, index=False)
        diversity = inverse_simpson(df_clono["umis"])
        n_clonotypes = len(df_clono)
        return n_clonotypes, diversity

    @utils.add_log
    def create_clonotypes(self):
        for barcode, sample in self.barcode_sample.items():
            annotation_file = f"{self.annotation_dir}/{sample}_annotation.csv"
            out_file = f"{self.clonotypes_dir}/{sample}_clonotypes.csv"
            n_clonotypes, diversity = self.create_well_clonotypes(
                annotation_file, out_file
            )
            self.metrics[barcode]["n_clonotypes"] = n_clonotypes
            self.metrics[barcode]["diversity"] = round(diversity, 2)

    @utils.add_log
    def merge_files(self):
        annotation_files = [
            f"{self.annotation_dir}/{sample}_annotation.csv"
            for sample in self.barcode_sample.values()
        ]
        utils.merge_table_files(annotation_files, self.annotation_csv)

        clonotypes_files = [
            f"{self.clonotypes_dir}/{sample}_clonotypes.csv"
            for sample in self.barcode_sample.values()
        ]
        utils.merge_table_files(clonotypes_files, self.clonotypes_csv)

    @utils.add_log
    def add_metrics(self):
        metrics_df = pd.DataFrame.from_dict(self.metrics, orient="index")
        # add percent
        umi_cols = [
            "umi_mapped",
            "umi_confident",
        ]
        umi_cols.extend([f"umi_confident_{chain}" for chain in self.chains])
        for col in umi_cols:
            metrics_df[col] = metrics_df.apply(
                lambda row: f"{row[col]}({round(row[col]/row['umi']*100, 2)}%)", axis=1
            )

        metrics_df.sort_values("read", inplace=True)
        metrics_df.to_csv(f"{self.out_prefix}_well_metrics.tsv", sep="\t")
        well_metrics_df = metrics_df[
            metrics_df["sample"].isin(self.barcode_sample.values())
        ]
        well_reads = well_metrics_df["read"].sum()
        total_reads = metrics_df["read"].sum()

        self.add_metric(
            name="Fraction of Reads in Wells",
            value=well_reads / total_reads,
            value_type="fraction",
            help_info="Lower value indicates wrong well_sample file or high ambient contamination.",
        )
        well_metrics_df.sort_values("well", inplace=True)

        cols = ["well", "sample", "n_clonotypes", "diversity"]
        well_metrics_df = well_metrics_df[
            cols + [col for col in well_metrics_df.columns if col not in cols]
        ]
        self.add_table(
            title="Well Metrics",
            table_id="well_metrics",
            df=well_metrics_df,
            help="n_clonotypes: total number of clonotypes<\br>diversity: inverse Simpson diversity index<\br>",
        )

        df = pd.read_csv(self.clonotypes_csv, sep=",")
        df = df.drop(["barcode"], axis=1)
        df = df.groupby("sample", group_keys=False).head(10)
        self.add_table(
            title="Top 10 Clonotypes",
            table_id="clonotypes",
            df=df,
        )

    def run(self):
        self.igblast()
        self.process_airr()
        self.create_clonotypes()
        self.merge_files()
        self.add_metrics()


def mapping_vdj(args):
    with Mapping_vdj(args, display_title="Mapping") as runner:
        runner.run()


def get_opts_mapping_vdj(parser, sub_program):
    parser.add_argument(
        "--well_sample",
        help="tsv file of well numbers and sample names. The first column is well numbers and the second column is sample names.",
        required=True,
    )
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
        - `C`: cell barcode  
        - `L`: linker(common sequences)  
        - `U`: UMI    
        - `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    super_vdj.get_opts_mapping_vdj(parser, sub_program)
    if sub_program:
        parser.add_argument("--consensus_metrics_file", required=True)



STEPS = [
    "mkref",
    "sample",
    "barcode",
    "cutadapt",
    "consensus",
    "star",
    "featureCounts",
    "target_metrics",
    "variant_calling",
    "filter_snp",
    "analysis_snp",
]
__ASSAY__ = "snp"
IMPORT_DICT = {"star": "celescope.rna"}

PANEL = {"lung_1", "blood_1", "CHIP"}



from celescope.rna_5p3p.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_rna_5p3p(Multi):
    """
    ## Usage
    ```
        multi_rna\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --thread 8\\
        --mod shell
    ```
    Work for both single cell RNA-Seq and single nuclei RNA-Seq.
    """

    def get_5p3p_fq(self, sample):
        arr = self.fq_dict[sample]
        arr["fq1_3p"] = []
        arr["fq2_3p"] = []
        arr["fq1_5p"] = []
        arr["fq2_5p"] = []
        for fq1, fq2, col4 in zip(arr["fq1"], arr["fq2"], arr["col4"]):
            if col4 not in ("5p", "3p"):
                raise Exception(
                    f"ERROR: The 4th column of mapfile must be 3p or 5p. But {col4} found."
                )
            arr[f"fq1_{col4}"].append(fq1)
            arr[f"fq2_{col4}"].append(fq2)
        return arr

    def convert(self, sample):
        step = "convert"
        arr = self.get_5p3p_fq(sample)
        cmd_line = self.get_cmd_line(step, sample)
        fq1_5p = ",".join(arr["fq1_5p"])
        fq1_3p = ",".join(arr["fq1_3p"])
        fq2_5p = ",".join(arr["fq2_5p"])
        fq2_3p = ",".join(arr["fq2_3p"])
        cmd = (
            f"{cmd_line} "
            f"--fq1_5p {fq1_5p} --fq1_3p {fq1_3p} "
            f"--fq2_5p {fq2_5p} --fq2_3p {fq2_3p} "
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def starsolo(self, sample):
        step = "starsolo"
        arr = self.get_5p3p_fq(sample)
        cmd_line = self.get_cmd_line(step, sample)
        fq1_list, fq2_list = [], []
        cnt = {"3p": 0, "5p": 0}
        for x in arr["col4"]:
            cnt[x] += 1
            fq1 = f'{self.outdir_dic[sample]["convert"]}/{sample}_{x}{cnt[x]}_R1.fq.gz'
            fq2 = f'{self.outdir_dic[sample]["convert"]}/{sample}_{x}{cnt[x]}_R2.fq.gz'
            fq1_list.append(fq1)
            fq2_list.append(fq2)
        fq1_str = ",".join(fq1_list)
        fq2_str = ",".join(fq2_list)
        cmd = (
            f"{cmd_line} "
            f"--fq1 {fq1_str} --fq2 {fq2_str} "
            f'--soloFeatures "GeneFull_Ex50pAS Gene SJ" '
        )
        self.process_cmd(
            cmd,
            step,
            sample,
            m=int(self.args.limitBAMsortRAM / 1e9),
            x=self.args.thread,
        )

    def analysis(self, sample):
        step = "analysis"
        matrix_file = f'{self.outdir_dic[sample]["outs"]}/filtered'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = f"{cmd_line} " f"--matrix_file {matrix_file} "
        self.process_cmd(cmd, step, sample, m=10, x=1)


def main():
    multi = Multi_rna_5p3p(__ASSAY__)
    multi.run()


if __name__ == "__main__":
    main()



import abc
import sys
import io
import json
import numbers
import os
import subprocess

from jinja2 import Environment, FileSystemLoader, select_autoescape

from celescope.tools import utils
from celescope.__init__ import HELP_DICT, __version__


def cap_str_except_preposition(my_string):
    prepositions = {
        "and",
        "or",
        "the",
        "a",
        "of",
        "in",
        "per",
        "after",
        "with",
        "by",
        "at",
    }
    lowercase_words = my_string.split(" ")

    final_words = [
        word if word in prepositions else word[0].upper() + word[1:]
        for word in lowercase_words
    ]
    final_words = " ".join(final_words)
    return final_words


def s_common(parser):
    """subparser common arguments"""
    parser.add_argument("--outdir", help="Output diretory.", required=True)
    parser.add_argument("--sample", help="Sample name.", required=True)
    parser.add_argument("--thread", help=HELP_DICT["thread"], default=4)
    parser.add_argument("--debug", help=HELP_DICT["debug"], action="store_true")
    return parser


class Step:
    """
    Step class
    """

    def __init__(self, args, display_title=None):
        """
        display_title controls the section title in HTML report
        force thread <=20
        """
        sys.stderr.write(f"CeleScope version: {__version__} ")
        sys.stderr.write(f"Args: {args}\n")
        self.args = args
        self.outdir = args.outdir
        self.outs_dir = f"{args.outdir}/../outs"
        self.sample = args.sample
        self.assay = args.subparser_assay
        self.thread = int(args.thread)
        self.thread = min(self.thread, 20)
        self.debug = args.debug
        self.out_prefix = f"{self.outdir}/{self.sample}"
        self.display_title = display_title

        # metrics index
        self.metric_index = 0

        utils.check_mkdir(self.outdir)

        # set
        class_name = self.__class__.__name__
        if not display_title:
            self._display_title = class_name
        else:
            self._display_title = display_title
        self._step_name = class_name[0].lower() + class_name[1:]
        self.__slots = ["data", "metrics"]
        self._step_summary_name = f"{self._step_name}_summary"

        self.__metric_list = []
        self.__help_content = []
        self.__comments = []
        self._path_dict = {}
        for slot in self.__slots:
            self._path_dict[slot] = f"{self.outdir}/../.{slot}.json"

        self.__content_dict = {}
        self.old_step_dict = {}
        for slot, path in self._path_dict.items():
            if not os.path.exists(path):
                self.__content_dict[slot] = {}
                if slot == "data":
                    self.__content_dict[slot]["parameters"] = {}
            else:
                with open(path) as f:
                    try:
                        self.__content_dict[slot] = json.load(f)
                    except ValueError:
                        print(
                            f'WARNING: Decoding "{path}" as json has failed. Will create empty json file.'
                        )
                        self.__content_dict[slot] = {}
            # clear step_summary
            if self._step_summary_name in self.__content_dict[slot]:
                self.old_step_dict[slot] = self.__content_dict[slot][
                    self._step_summary_name
                ]
            self.__content_dict[slot][self._step_summary_name] = {}

        # jinja env
        self.env = Environment(
            loader=FileSystemLoader(os.path.dirname(__file__) + "/../templates/"),
            autoescape=select_autoescape(["html", "xml"]),
        )

        # out file
        self.__stat_file = f"{self.outdir}/stat.txt"
        self.report_html = f"{self.outdir}/../{self.sample}_report.html"

        # move file to outs
        self.outs = []

    def add_metric(
        self,
        name,
        value,
        total=None,
        help_info=None,
        display=None,
        show=True,
        print_log=True,
        value_type=None,
    ):
        """
        add metric to metric_list

        Args
            total: int or float, used to calculate fraction
            help_info: str, help info for metric in html report
            display: str, controls how to display the metric in HTML report.
            show: bool, whether to add to `.data.json` and `stat.txt`. `.data.json` is used for HTML report. `stat.txt` is used in house.
            print_log: bool, whether to print metric to stdout
            value_type: if value_type is "fraction", value = round(value * 100, 2), display = f'{value}%'
        """

        name = cap_str_except_preposition(name)
        if help_info:
            help_info = help_info[0].upper() + help_info[1:]
            if help_info[-1] != ".":
                help_info += "."
        if not display:
            if isinstance(value, numbers.Number):
                display = str(format(value, ","))
            else:
                display = value
        fraction = None
        if total:
            fraction = round(value / total * 100, 2)
            display += f"({fraction}%)"
        if value_type == "fraction":
            value = round(value * 100, 2)
            display = f"{value}%"
        self.__metric_list.append(
            {
                "name": name,
                "value": value,
                "total": total,
                "fraction": fraction,
                "display": display,
                "help_info": help_info,
                "show": show,
            }
        )

        if print_log:
            sys.stderr.write(f"{name}: {display}\n")

    def _write_stat(self):
        with open(self.__stat_file, "w") as writer:
            for metric in self.__metric_list:
                if metric["show"]:
                    name = metric["name"]
                    display = metric["display"]

                    line = f"{name}: {display}"
                    writer.write(line + "\n")

    def _dump_content(self):
        """dump content to json file"""
        for slot, path in self._path_dict.items():
            if self.__content_dict[slot]:
                with open(path, "w") as f:
                    json.dump(self.__content_dict[slot], f, indent=4)

    @utils.add_log
    def _render_html(self):
        template = self.env.get_template(f"html/{self.assay}/base.html")
        with io.open(self.report_html, "w", encoding="utf8") as f:
            html = template.render(self.__content_dict["data"])
            f.write(html)

    def _add_content_data(self):
        step_summary = {}
        step_summary["display_title"] = self._display_title
        metric_list = []
        comment_metric_list = []
        for metric in self.__metric_list:
            if metric["show"]:
                metric_list.append(metric)
            else:
                comment_metric_list.append(metric)
        step_summary["metric_list"] = metric_list
        step_summary["comment_metric_list"] = comment_metric_list
        step_summary["help_content"] = self.__help_content
        step_summary["comments"] = self.__comments
        self.__content_dict["data"][self._step_summary_name].update(step_summary)

    def _add_content_metric(self):
        metric_dict = dict()
        for metric in self.__metric_list:
            name = metric["name"]
            value = metric["value"]
            fraction = metric["fraction"]
            metric_dict[name] = value
            if fraction:
                metric_dict[f"{name} Fraction"] = fraction

        self.__content_dict["metrics"][self._step_summary_name].update(metric_dict)

    def add_data(self, **kwargs):
        """
        add data(other than metrics) to self.content_dict['data']
        for example: add plots and tables
        """
        for key, value in kwargs.items():
            self.__content_dict["data"][self._step_summary_name][key] = value

    def add_help_content(self, name, content):
        """
        add help info before metrics' help_info
        """
        self.__help_content.append({"name": name, "content": content})

    def add_comments(self, content):
        """
        add comment to self.__comments
        """
        self.__comments.append(content)

    @utils.add_log
    def get_slot_key(self, slot, step_name, key):
        """read slot from json file"""
        try:
            return self.__content_dict[slot][step_name + "_summary"][key]
        except KeyError:
            self.get_slot_key.logger.warning(
                f"{key} not found in {step_name}_summary.{slot}"
            )
            raise

    @utils.add_log
    def add_slot_step(self, slot, step_name, val):
        """add slot to json"""
        self.__content_dict[slot][step_name + "_summary"] = val

    @utils.add_log
    def get_slot_step(self, slot, step_name):
        """add slot to json"""
        return self.__content_dict[slot][step_name + "_summary"]

    @utils.add_log
    def add_table(self, title, table_id, df, help="", script=""):
        if not script:
            script = """
            <script>
            $(document).ready(function () {{
                var table = $('#{table_id}').DataTable({{
                    "order": [],
                    "scrollX": true,
                    "autoWidth": false,
                    dom: 'Bfrtip',
                    buttons: ['excel']
                }});
                table.draw();
            }});
            </script>
            """.format(table_id=table_id)

        self.__content_dict["data"][table_id] = {
            "title": title,
            "df": df.to_html(
                escape=False, index=False, table_id=table_id, justify="center"
            ),
            "help": help,
            "script": script,
        }

    def get_table_dict(self, title, table_id, df_table):
        """
        table_dict {title: '', table_id: '', df_table: pd.DataFrame}
        """
        table_dict = {}
        table_dict["title"] = title
        table_dict["table"] = df_table.to_html(
            escape=False, index=False, table_id=table_id, justify="center"
        )
        table_dict["id"] = table_id
        return table_dict

    def _move_files(self):
        if self.outs:
            utils.check_mkdir(self.outs_dir)
        for f in self.outs:
            cmd = ""
            if not os.path.exists(f):
                sys.stderr.write(
                    f"WARNING: output file {f} not found! The pipeline may have failed.\n"
                )
            elif os.path.isfile(f):
                cmd = f"mv -f {f} {self.outs_dir}"
            elif os.path.isdir(f):
                cmd = f"set -e; cp -r {f} {self.outs_dir}; "
            subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def _add_parameters(self):
        for key, value in vars(self.args).items():
            if key not in {
                "func",
                "thread",
                "outdir",
                "sample",
                "subparser_assay",
                "debug",
            }:
                self.__content_dict["data"]["parameters"][key] = value

    @utils.add_log
    def _clean_up(self):
        self._add_content_data()
        self._add_content_metric()
        self._add_parameters()
        self._write_stat()
        self._dump_content()
        self._render_html()
        self._move_files()

    @utils.add_log
    def debug_subprocess_call(self, cmd):
        """
        debug subprocess call
        """
        self.debug_subprocess_call.logger.info(cmd)
        if cmd.find("2>&1") == -1:
            cmd += " 2>&1 "
        subprocess.check_call(cmd, shell=True)

    def get_metric_list(self):
        return self.__metric_list

    def set_metric_list(self, metric_list):
        self.__metric_list = metric_list

    @abc.abstractmethod
    def run(self):
        sys.exit("Please implement run() method.")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self._clean_up()



import os
import celescope.tools


STEPS = ["sample", "barcode", "mapping", "assemble", "summarize", "annotation"]
__ASSAY__ = "flv_trust4"

TOOLS_DIR = os.path.dirname(celescope.tools.__file__) + "/trust4"
REF_DIR = f"{TOOLS_DIR}/database"

CHAIN = {"TCR": ["TRA", "TRB"], "BCR": ["IGH", "IGL", "IGK"]}
PAIRED_CHAIN = {"TCR": ["TRA_TRB"], "BCR": ["IGK_IGH", "IGL_IGH"]}



import argparse
import sys
import subprocess
import os

from celescope.tools.step import Step
from celescope.tools.barcode import get_opts_barcode
from celescope.tools.cutadapt import get_opts_cutadapt, get_cutadapt_cmd, Cutadapt as Ct
from celescope.rna.star import get_opts_star
from celescope.tools.star_mixin import get_star_cmd, Star_mixin
from celescope.tools import utils


def args_to_list(args):
    s = []
    for arg, v in vars(args).items():
        if v not in (False, None, "") and arg != "func":
            s.append(f"--{arg}")
            if v is not True:
                if "param" in arg:
                    s.append(f'"{v}"')
                else:
                    s.append(f"{v}")
    return s


def get_step_args_str(args, get_opts_func):
    args_str = ""
    parser = argparse.ArgumentParser()
    get_opts_func(parser, sub_program=True)
    # [known_args_namespace, not_known_args_list]
    known_args = parser.parse_known_args(args_to_list(args))[0]
    args_str = " ".join(args_to_list(known_args))
    return args_str


class Prep_map(Step):
    """
    ## Features
    - `prep_map` contains preprocessing (barcode, cutadapt) and mapping steps. The output bam file will be used for downstream quantitative analysis.

    ## Output
    - `{sample}_Aligned.sortedByCoord.out.bam` Bam file sort by coordinates.
    """

    def __init__(self, args):
        super().__init__(args)
        self.bc2 = f"{self.out_prefix}_bc2.FIFO"
        self.ct2 = f"{self.out_prefix}_ct2.FIFO"

    @utils.add_log
    def run(self):
        bc2, ct2 = self.bc2, self.ct2
        bc_args_str = get_step_args_str(self.args, get_opts_barcode)
        ct_cmd = get_cutadapt_cmd(self.args, bc2, ct2)
        star_cmd = get_star_cmd(self.args, ct2, self.out_prefix)
        run_assay = "rna"
        if self.assay == "bulk_rna":
            run_assay = self.assay
        cmd = (
            f"mkfifo {bc2} {ct2} \n"
            f"celescope {run_assay} barcode {bc_args_str} --stdout >> {bc2} & \n"
            f"{ct_cmd} & \n"
            f"{star_cmd} \n"
            f"rm {bc2} {ct2} \n"
        )
        sys.stderr.write(cmd + "\n")
        subprocess.check_call(cmd, shell=True)

    def __exit__(self, *args, **kwargs):
        """
        make sure fifo are removed
        """
        for fifo in (self.bc2, self.ct2):
            if os.path.exists(fifo):
                os.remove(fifo)


class Cutadapt(Ct):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def run(self):
        self.add_cutadapt_metrics()


class Star(Star_mixin):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def run(self):
        self.add_star_metrics()


@utils.add_log
def prep_map(args):
    with Prep_map(args) as runner:
        runner.run()

    with Cutadapt(args, display_title="Trimming") as runner:
        runner.run()

    with Star(args, display_title="Mapping") as runner:
        runner.run()


def get_opts_prep_map(parser, sub_program):
    get_opts_barcode(parser, sub_program=sub_program)
    get_opts_cutadapt(parser, sub_program=False)
    get_opts_star(parser, sub_program=False)



from celescope.tools import utils
from celescope.__init__ import __VERSION__
from celescope.chemistry_dict import chemistry_dict

from celescope.tools.step import Step, s_common

from celescope.tools.parse_chemistry import get_chemistry


class Sample(Step):
    def __init__(self, args):
        Step.__init__(self, args)
        self.fq1_list = args.fq1.split(",")

    @utils.add_log
    def run(self):
        chemistry = get_chemistry(self.assay, self.args.chemistry, self.fq1_list)

        self.add_metric(
            name="Sample ID",
            value=self.sample,
        )
        self.add_metric(
            name="Assay",
            value=self.assay,
        )
        self.add_metric(
            name="Chemistry",
            value=chemistry,
            help_info='For more information, see <a href="https://github.com/singleron-RD/CeleScope/blob/master/doc/chemistry.md">here</a>',
        )
        self.add_metric(
            name="Software Version",
            value=__VERSION__,
        )


@utils.add_log
def sample(args):
    with Sample(args) as runner:
        runner.run()


def get_opts_sample(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--fq1", help="read1 fq file")
    parser.add_argument(
        "--chemistry",
        choices=list(chemistry_dict.keys()),
        help="chemistry version",
        default="auto",
    )
    return parser



from celescope.bulk_rna.__init__ import __ASSAY__
from celescope.tools.multi import Multi
from celescope.tools.__init__ import (
    OUTS_DIR,
)


class Multi_bulk_rna(Multi):
    def starsolo(self, sample):
        step = "starsolo"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = f'{cmd_line} ' f'--fq1 {arr["fq1_str"]} --fq2 {arr["fq2_str"]} '
        self.process_cmd(
            cmd,
            step,
            sample,
            m=int(self.args.limitBAMsortRAM / 1e9),
            x=self.args.thread,
        )

    def split(self, sample):
        step = "split"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr["fq1_str"]} '
            f'--bam {self.outdir_dic[sample][OUTS_DIR]}/{sample}_Aligned.sortedByCoord.out.bam '
            f'--barcodes {self.outdir_dic[sample][OUTS_DIR]}/filtered/barcodes.tsv.gz '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)


def main():
    multi = Multi_bulk_rna(__ASSAY__)
    multi.run()


if __name__ == "__main__":
    main()
